#define F_CPU 16000000UL

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>
#include <avr/pgmspace.h>

#include "tft.h"
#include "font.h"

#define LCD_RESET 0

volatile unsigned int X_SIZE = 0;	//   
volatile unsigned int Y_SIZE = 0;	//   
unsigned long XY_LEN = 76800UL;

#define LCD_C (*(volatile uint8_t*)(0b1000000000000000))     // 32768
#define LCD_D (*(volatile uint8_t*)(0b1000100000000000))     // 34816

void TFT9341_SendCommand(unsigned char cmd)
{	LCD_C = cmd;
}

void TFT9341_SendData(unsigned char data)
{	LCD_D = data;
}

void TFT9341_reset(void)
{	DDRE |= (1<<LCD_RESET);
 	PORTE &= ~(1<<LCD_RESET);
  	_delay_ms(100);
 	PORTE |= (1<<LCD_RESET);
}

inline void TFT9341_WriteRegister32(unsigned char r, unsigned long d)
{	LCD_C = r;
	LCD_D = (d>>24);
	LCD_D = (d>>16);
	LCD_D = (d>>8);
	LCD_D = (d);
}

void TFT9341_SetAddrWindow(unsigned int x1,unsigned int y1,unsigned int x2,unsigned int y2)
{	unsigned long t = x1;
  	t <<= 16;
  	t |= x2;
	LCD_C = 0x2A;
	LCD_D = (t>>24);
	LCD_D = (t>>16);
	LCD_D = (t>>8);
	LCD_D = (t);

  	t = y1;
  	t <<= 16;
  	t |= y2;
	LCD_C = 0x2B;
	LCD_D = (t>>24);
	LCD_D = (t>>16);
	LCD_D = (t>>8);
	LCD_D = (t);
}

void TFT9341_SetRotation(unsigned char r)
{  	TFT9341_SendCommand(0x36);
	switch(r)
  	{  	case 0:
			TFT9341_SendData(0x48);
			X_SIZE = 240;
			Y_SIZE = 320;
    		break;
    	case 1:
			TFT9341_SendData(0x28);
			X_SIZE = 320;
			Y_SIZE = 240;
    		break;
    	case 2:
			TFT9341_SendData(0x88);
			X_SIZE = 240;
			Y_SIZE = 320;
    		break;
    	case 3:
			TFT9341_SendData(0xE8);
			X_SIZE = 320;
			Y_SIZE = 240;
    		break;
  	}
}

unsigned int Color565(unsigned char r, unsigned char g, unsigned char b)
{ 	unsigned int c = 0;
  	c = r >> 3;
  	c <<= 6;
  	c |= g >> 2;
  	c <<= 5;
  	c |= b >> 3;
  	return c;
}

void TFT9341_DrawPixel(
	int x, //  
	int y, //  
	unsigned int color)
{
//    
	if((x<0)||(y<0)||(x>=X_SIZE)||(y>=Y_SIZE)) return;
  
  	TFT9341_SetAddrWindow(x,y,X_SIZE-1,Y_SIZE-1);
	LCD_C = 0x2C;
	LCD_D = (color>>8);
	LCD_D = color;
}

void TFT9341_DrawLine(
	unsigned int color,
	unsigned int x1, unsigned int y1, 
	unsigned int x2, unsigned int y2)
{
	int steep = abs(y2 - y1) > abs(x2 - x1);
	if (steep)
	{	swap(x1, y1);
  		swap(x2, y2);
	}
	if (x1 > x2)
	{	swap(x1, x2);
  		swap(y1, y2);
	}
 
	int dx = (x2 - x1), 
		dy = abs(y2 - y1);
 
	int err = dx / 2;
	int ystep;
	if (y1 < y2)
	{
  		ystep = 1;
	} 
	else
	{
  		ystep = -1;
	}

  	for (; x1 <= x2; x1++)
	{
 		if (steep)
		{
  			TFT9341_DrawPixel(y1, x1, color);
		} 
		else 
		{
  			TFT9341_DrawPixel(x1, y1, color);
		}

   		err -= dy;
     	if (err < 0)
		{
      		y1 += ystep;
      		err += dx;
    	}
	}
}

void TFT9341_Flood(unsigned short color, unsigned long len)
{
  	unsigned short blocks;
  	unsigned char i, hi = color, lo=(color>>8);

	LCD_C = 0x2C;
	LCD_D = hi; LCD_D = lo;
 
	LCD_D = lo;
	len--;

	blocks = (unsigned short)(len/64);//64 pixels/block
 
	while(blocks--)
	{	i = 16;
		do
		{	LCD_D = hi; LCD_D = lo; 
			LCD_D = hi; LCD_D = lo; 
			LCD_D = hi; LCD_D = lo; 
			LCD_D = hi; LCD_D = lo; 
		} while (--i);
	}

  	for (i = (unsigned char)len & 63; i--;)
  	{	LCD_D = hi; LCD_D = lo; 
  	}
}

void TFT9341_FillScreen(unsigned int color)
{
  	TFT9341_SetAddrWindow(0,0,X_SIZE-1,Y_SIZE-1);
  	TFT9341_Flood(color,XY_LEN);
}


void TFT9341_FillRectangle(
	unsigned int color,					// 
	unsigned int x1, unsigned int y1,	//   
	unsigned int x2, unsigned int y2)
{
	TFT9341_SetAddrWindow(x1, y1, x2, y2);
	TFT9341_Flood(color, (long)(x2-x1+1) * (long)(y2-y1+1));
}

void TFT9341_Draw_Char(	//    1,2,3,4
	int x,			//  
	int y, 			//  
	unsigned int color, unsigned int backcolor, 
	unsigned char ch, 
	unsigned char size)
{
	unsigned char value = 0;
	unsigned char mask[8] =
	{	0b10000000,
		0b01000000,
		0b00100000,
		0b00010000,
		0b00001000,
		0b00000100,
		0b00000010,
		0b00000001
	};

 	if (ch > 127) ch = ch-64;

 	switch(size)
  	{
    	int v,h;
    	int tv,th;
    	int tcolor;

		case 1:
  			for(h = 0; h < 5; h++)
  			{	value = pgm_read_byte(&(font5x8[ch-32][h]));
				for(v = 0; v < 8; v++)
				{	if ((value & mask[v]) != 0) tcolor = color;
					else tcolor = backcolor;
					TFT9341_DrawPixel(x+v,y+h,tcolor);
				}
  			}
  			break;
 
    	case 2:
  			for(h = 0; h < 5; h++)
  			{	value = pgm_read_byte(&(font5x8[ch-32][h]));
				th = 2*h;
    			for(v = 0; v < 8; v++)			//   (x)
				{	tv = 2*v;
					if ((value & mask[v]) != 0) tcolor = color;
					else tcolor = backcolor;
					TFT9341_DrawPixel(x + tv,		y + th,		tcolor);
					TFT9341_DrawPixel(x + tv,		y + (th+1),	tcolor);
					TFT9341_DrawPixel(x + (tv+1),	y + th,		tcolor);
					TFT9341_DrawPixel(x + (tv+1),	y + (th+1),	tcolor);
  				}
			}
    		break;
	}
}

void print_str(
		unsigned char x_pos, unsigned char y_pos,
		unsigned char font_size,
		unsigned int colour, unsigned int back_colour,
		char *str)
{	while (*str)
	{	TFT9341_Draw_Char(x_pos,y_pos,colour,back_colour,*str++,font_size);
		y_pos += (font_size * 5);
	}
}
